%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% IEEE CEC/GECCO 2023 competition: Competition on Evolutionary Computation in the Energy Domain: Operation and Planning Applications
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% ALGORITMH: CHAOTIC CAUCHY HC2RCEDUMDA (CC_HC2RCEDUMDA)
% CHAOTIC CAUCHY HC2RCEDUMDA
%This algorithm is the modified version of the HC2RCEDUMDA using CHAOTIC CAUCHY distribution.
% The codes have been modified by the developers to propose CC_HC2RCEDUMDA
%% Developers: 
% Main author: Dharmesh A. Dabhi, POST DOCTORAL RESEARCHER, TECHNOLOGICAL UNIVERSITY DUBLIN,DUBLIN
% Co-authors: 
% Kartik S. Pandya,Independent Researcher
% Pratik Mochi, Assistant professor, M & V Patel Department of Electrical Engineering, CSPIT,CHARUSAT UNIVERSITY,CHANGA, Gujarat, INDIA
% ROHIT SALGOTRA, FACULTY OF PHYSICS & APPLIED COMPUTER SCIENCE, AGH UNIVERSITY OF SCIENCE & TECHNOLOGY, POLAND.
%% References:
% The codes of HC2RCEDUMDA available http://www.gecad.isep.ipp.pt/ERM-competitions/2021-2/
% The Codes of HC2RCEDUMDA were developed by the following team in IEEE PES GM 2021, IEEE %CEC 2021 & GECCO 2021 Joint competition 2021
% Authors: Ansel Y. Rodrguez Gonzlez 1,2 (Team Leader) 
%          Samantha Barajas            1,3
%          Ramn Aranda                1,2
%          Yoan Martnez Lpez         4
%          Julio Madera                4


function [Fit_and_p,FVr_bestmemit, fitMaxVector,evalCount,gen,enCodedBestSol] = ...
    hcd_extra(CC_HC2RCEDUMDAparameters,caseStudyData,otherParameters,low_habitat_limit,up_habitat_limit, codes, Fit_and_p, FVr_bestmemit, fitMaxVector, gen, evalCount, I_repeatEvals, FullStoreSol, FullStoreVal, FullStoreCount)

    maxEvals      = CC_HC2RCEDUMDAparameters.maxEvals;
    codeSize      = CC_HC2RCEDUMDAparameters.I_CodeSize;
    popSize       = 1;
    generations   = CC_HC2RCEDUMDAparameters.I_initHC_itermax;
    
    ratioFactor   = CC_HC2RCEDUMDAparameters.I_initHC_ratioFactor;
    maxStep       = floor((CC_HC2RCEDUMDAparameters.I_CodeSize-1) * ratioFactor);
    if (maxStep<1)
        maxStep = 1;
    end
        
    bestSol = FVr_bestmemit;
    bestVal = Fit_and_p;
    
    enCodedBestSol = enCode(bestSol,codes,low_habitat_limit,up_habitat_limit);
    
    genCount = 0;
       
    while (evalCount < maxEvals - popSize - 1) && (genCount<generations)
        
        enCodedPop = generateNewPop(enCodedBestSol,popSize,maxStep, codeSize, low_habitat_limit,up_habitat_limit);        
        deCodedPop = deCodePop(enCodedPop,codes);
      
        fitMaxVector(1,gen) = bestVal;        
        
        genCount = genCount + 1;
    end
   
   FVr_bestmemit = bestSol;
   Fit_and_p=[fitMaxVector(1,gen) 0];
   
function Pop = generateNewPop(enCodedBestSol,popSize, maxStep, codeSize, low_habitat_limit,up_habitat_limit)
    [~,cws] = size(enCodedBestSol);        
    Pop = randi([-maxStep maxStep],popSize,cws);

    for j=1:cws       
       if (low_habitat_limit(j)==up_habitat_limit(j))
            Pop(1,j) = 1;
       else
           Pop(1,j) =  enCodedBestSol(j) + Pop(1,j);
           
           if (Pop(1,j) < 1)
                Pop(1,j) = 1;
            end 
            if (Pop(1,j) > codeSize) 
                Pop(1,j) = codeSize;
            end
       end
    end
end    
    
% function [Val, evalCount] = my_feval(fnc,deCodedPop,caseStudyData, otherParameters, evalCount, I_repeatEvals)    
%     switch I_repeatEvals
%         case 0
%             %%%% Removed method. It is not relevant for HC2RCEDUMDA
%             % [Val, evalCount] = my_feval_withStore(fnc,deCodedPop,caseStudyData, otherParameters, evalCount);
%         case 2
%            [Val, evalCount] = my_feval_withStoreForElitism(fnc,deCodedPop,caseStudyData, otherParameters, evalCount); 
%         otherwise    
%             [s,~] = size(deCodedPop);    
%             evalCount = evalCount + s;
%             [Val, ~] = feval(fnc,deCodedPop,caseStudyData, otherParameters);
%             fprintf('evalCount: %d\n',evalCount);       
%     end
% end  

function [Val, evalCount] = my_feval_withStoreForElitism(fnc,deCodedPop,caseStudyData, otherParameters, evalCount)    
     [cs,ws] = size(deCodedPop);
     Val    = nan(1,cs);
     
     for iii=1:cs
        index = searchSolution(deCodedPop(iii,:),ws);
        if index > 0
            Val(iii) = FullStoreVal(index);           
        else
           auxPop = deCodedPop(iii,:);
           [auxVal, ~] = feval(fnc,auxPop,caseStudyData, otherParameters);
           Val(iii) = auxVal;
           evalCount = evalCount + 1;          
        end
     end    

     fprintf('evalCount: %d\n',evalCount);
end

function index = searchSolution(sol, ws)
     index = 0; 
    
     for iii=1:FullStoreCount
                 
         flag = 1;
         for jjj=1:ws
            if FullStoreSol(iii,jjj)~=sol(jjj)
              flag = 0;
              break;
            end
         end
         
         if flag == 1
            index = iii;
            fprintf('*');
            break; 
         end         
     end
end       

 function deCodedPop = deCodePop(enCodedPop,codes)
    [hs,ws] = size(enCodedPop);
    deCodedPop = nan(hs,ws);
    for i=1:hs
       deCodedPop(i,:) = deCode(enCodedPop(i,:),codes);
    end 
 end
    
 function indiv = deCode(enCoded,codes)
   [~,ws] = size(codes);
    indiv = nan(1,ws);   
    for j=1:ws
        pos = enCoded(1,j);
        v   = codes(pos,j);
        indiv(1,j)=v;
    end
 end

function enCoded = enCode(indiv,codes,FVr_minbound,FVr_maxbound)
    [hs,ws] = size(codes);
    enCoded = nan(1,ws);
    for jjj=1:ws
       if (FVr_minbound(jjj)==FVr_maxbound(jjj)) 
            enCoded(1,jjj)=1;
       else           
            range = (FVr_maxbound(jjj) - FVr_minbound(jjj))/hs;
            sum = FVr_minbound(jjj) + range;
            iii = 1;
            while (indiv(jjj) > sum) && (iii<hs) 
                  iii = iii + 1;
                  sum = sum + range;
            end
            enCoded(1,jjj)=iii;
       end
    end
end 
 
end
