%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% IEEE CEC/GECCO 2023 competition: Competition on Evolutionary Computation in the Energy Domain: Operation and Planning Applications
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% ALGORITMH: CHAOTIC LEAVY HC2RCEDUMDA (CL_HC2RCEDUMDA)
% CHAOTIC LEAVY HC2RCEDUMDA
%This algorithm is the modified version of the HC2RCEDUMDA using CHAOTIC LEAVY distribution.
% The codes have been modified by the developers to propose CL_HC2RCEDUMDA
%% Developers: 
% Main author: Dharmesh A. Dabhi, POST DOCTORAL RESEARCHER, TECHNOLOGICAL UNIVERSITY DUBLIN,DUBLIN
% Co-authors: 
% Kartik S. Pandya,Independent Researcher
% Pratik Mochi, Assistant professor, M & V Patel Department of Electrical Engineering, CSPIT,CHARUSAT UNIVERSITY,CHANGA, Gujarat, INDIA
% ROHIT SALGOTRA, FACULTY OF PHYSICS & APPLIED COMPUTER SCIENCE, AGH UNIVERSITY OF SCIENCE & TECHNOLOGY, POLAND.
%% References:
% The codes of HC2RCEDUMDA available http://www.gecad.isep.ipp.pt/ERM-competitions/2021-2/
% The Codes of HC2RCEDUMDA were developed by the following team in IEEE PES GM 2021, IEEE %CEC 2021 & GECCO 2021 Joint competition 2021
% Authors: Ansel Y. Rodrguez Gonzlez 1,2 (Team Leader) 
%          Samantha Barajas            1,3
%          Ramn Aranda                1,2
%          Yoan Martnez Lpez         4
%          Julio Madera                4

function [Fit_and_p,FVr_bestmemit, fitMaxVector] = ...
    CL_HC2RCEDUMDA(CL_HC2RCEDUMDAparameters,caseStudyData,otherParameters,low_habitat_limit,up_habitat_limit)

I_NP         = CL_HC2RCEDUMDAparameters.I_NP;
I_D          = numel(up_habitat_limit);

FVr_minbound = low_habitat_limit;
FVr_maxbound = up_habitat_limit;
I_itermax    = CL_HC2RCEDUMDAparameters.I_itermax;

I_indByCell         = CL_HC2RCEDUMDAparameters.I_indByCell;
I_StructureWidth    = CL_HC2RCEDUMDAparameters.I_StructureWidth;
I_StructureHeight   = CL_HC2RCEDUMDAparameters.I_StructureHeight;
I_NeighborhoodType  = CL_HC2RCEDUMDAparameters.I_NeighborhoodType;
I_NeighborhoodParam = CL_HC2RCEDUMDAparameters.I_NeighborhoodParam;

I_SelectionMethod   = CL_HC2RCEDUMDAparameters.I_SelectionMethod;
I_Selected          = CL_HC2RCEDUMDAparameters.I_Selected; 

I_elitismParam = CL_HC2RCEDUMDAparameters.I_elitismParam;

I_CodeSize     = CL_HC2RCEDUMDAparameters.I_CodeSize;

I_EstimationParamsMethod = CL_HC2RCEDUMDAparameters.I_EstimationParamsMethod; 
I_EstimationParamsParam  = CL_HC2RCEDUMDAparameters.I_EstimationParamsParam;

I_EstimationStructureMethod = CL_HC2RCEDUMDAparameters.I_EstimationStructureMethod; 
I_EstimationStructureParam  = CL_HC2RCEDUMDAparameters.I_EstimationStructureParam ;

I_ScalerMethod = CL_HC2RCEDUMDAparameters.I_ScalerMethod; 
I_ScalerParam  = CL_HC2RCEDUMDAparameters.I_ScalerParam;

I_SamplingMethod = CL_HC2RCEDUMDAparameters.I_SamplingMethod; 
I_SamplingParam  = CL_HC2RCEDUMDAparameters.I_SamplingParam;

I_maxFits     = CL_HC2RCEDUMDAparameters.I_maxFits; 
I_repeatEvals = CL_HC2RCEDUMDAparameters.I_repeatEvals; 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
fnc          =  otherParameters.fnc;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

fitMaxVector = nan(1,I_itermax);

[~,solSize] = size(FVr_minbound); 

global FullStoreSol;  
global FullStoreVal;  
global FullStoreCount;

if (I_repeatEvals == 0)
    FullStoreSol   = zeros(I_NP * I_itermax,solSize);
    FullStoreVal   = zeros(1,I_NP * I_itermax);
    FullStoreCount = 0;
end 
if (I_repeatEvals == 2)
    FullStoreSol   = zeros(I_elitismParam,solSize);
    FullStoreVal   = zeros(1,I_elitismParam);
    FullStoreCount = 0;
end 

gen = 1; 
evalCount = 0;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%----FM_pop is a matrix of size I_NPx(I_D+1). It will be initialized------
%----with random values between the min and max values of the-------------
%----parameters-----------------------------------------------------------
% FLC modification - vectorization
minPositionsMatrix=repmat(FVr_minbound,I_NP,1);
maxPositionsMatrix=repmat(FVr_maxbound,I_NP,1);

%code inicialized
codes = initializeCodes(FVr_minbound,FVr_maxbound,I_CodeSize);

% generate initial population.
rand('state',otherParameters.iRuns) %Guarantee same initial population

% FM_pop=genpop(I_NP,I_D,minPositionsMatrix,maxPositionsMatrix);
FM_pop=minPositionsMatrix;
[S_val, evalCount] = my_feval(fnc,FM_pop,caseStudyData, otherParameters, evalCount,I_repeatEvals);   

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% HC 
%

if CL_HC2RCEDUMDAparameters.I_initHC_itermax>0
    
   [S_bestval,I_best_index] = min(S_val); 
   FVr_bestmemit = FM_pop(I_best_index,:);
   Fit_and_p     = S_bestval
   
   fitMaxVector(1,1) = Fit_and_p;
   
   [~,FVr_bestmemit, fitMaxVector,evalCount,gen, ~] = ...
    hcd_extra(CL_HC2RCEDUMDAparameters,caseStudyData,otherParameters,low_habitat_limit,up_habitat_limit, codes, Fit_and_p, FVr_bestmemit,fitMaxVector, gen, evalCount, I_repeatEvals, FullStoreSol, FullStoreVal, FullStoreCount);
 
   S_val(I_best_index) = fitMaxVector(1,gen); 
   FM_pop(I_best_index,:) = FVr_bestmemit;
     
end

 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  
FVr_bestmemit = FM_pop(I_best_index,:) 
fitMaxVector(1,gen)   = S_val(I_best_index);

enCodedPop = enCodePop(FM_pop,codes,FVr_minbound,FVr_maxbound);
newEnCodedPop = enCodedPop;

[~,pbidx_all] = sort(S_val);
pbidx = pbidx_all(1:I_elitismParam);    
pbpop = enCodedPop(pbidx,:);

 if (I_repeatEvals == 2)       
     FullStoreSol   = FM_pop(pbidx,:);
     FullStoreVal   = S_val(pbidx);
     FullStoreCount = I_elitismParam; 
 end

eCount = 0;

while (eCount<I_itermax) && (evalCount < CL_HC2RCEDUMDAparameters.maxEvals-I_NP)    
    for cellIndexH = 1:I_StructureHeight
        for cellIndexW = 1:I_StructureWidth
          
            NeighborsIndexes          = getNeighborsIndexes(cellIndexH,cellIndexW,I_StructureHeight,I_StructureWidth,I_NeighborhoodType,I_NeighborhoodParam,I_indByCell);
            SelectedIndividualIndexes = getSelectedIndividualIndexes(NeighborsIndexes,S_val,I_SelectionMethod,I_Selected);        
            
            [Px,Pxy] = getEstimadedParams(enCodedPop,SelectedIndividualIndexes,codes,I_EstimationParamsMethod,I_EstimationParamsParam);
           
            [Dxgy,EdgesList,EdgesCount] = getEstimadedStructure(Px,Pxy,I_EstimationStructureMethod,I_EstimationStructureParam);
            
            [Px,Pxy] = getScaledParams(Px,Pxy,length(SelectedIndividualIndexes),I_ScalerMethod,I_ScalerParam);
             
            cellLinearIndex = I_StructureWidth * (cellIndexH-1) + cellIndexW;
            CellIndividualIndexes = getCellIndividualLinearIndexes(cellLinearIndex, I_indByCell); 
                    
            NewCellPop = sampleNewIndividualsFromModel(I_indByCell,Px,Pxy,Dxgy,EdgesList,EdgesCount,I_SamplingMethod,I_SamplingParam);
          
            for ii = 1: I_indByCell            
                newEnCodedPop(CellIndividualIndexes(ii),:) = NewCellPop(ii,:);
            end
              
        end
    end
    
    enCodedPop = newEnCodedPop;
    
    %%% random idx to put k etitist individuals
    pwidx = randi([1 I_NP],1,I_elitismParam);

    enCodedPop(pwidx,:) = pbpop();
    deCodedPop = deCodePop(enCodedPop,codes);
      
    %Evaluation of new Pop
    [S_val, evalCount] = my_feval(fnc,deCodedPop,caseStudyData, otherParameters, evalCount,I_repeatEvals)
    
    [~,pbidx_all] = sort(S_val);
    pbidx = pbidx_all(1:I_elitismParam);    
    pbpop = enCodedPop(pbidx,:);

    if (I_repeatEvals == 2)       
        FullStoreSol   = deCodedPop(pbidx,:);
        FullStoreVal   = S_val(pbidx);
        FullStoreCount = I_elitismParam; 
    end
  
    avgV   = mean(S_val);
    worstV = max(S_val);
        
    %% update best results
    [S_bestval,I_best_index] = min(S_val)
    FVr_bestmemit = deCodedPop(I_best_index,:)
           
    fprintf('Best Worst Avg Fitness value: %f %f %f %f\n',S_val(I_best_index),S_bestval,avgV,worstV);
    fprintf('Generation: %d\n',gen);
 
    gen=gen+1; 
    
    fitMaxVector(1,gen) = S_val(I_best_index);
        
    eCount = eCount + 1;

end

Fit_and_p=[fitMaxVector(1,gen) 0]; 

[Fit_and_p,FVr_bestmemit, fitMaxVectorAux] = hcc_extra(CL_HC2RCEDUMDAparameters,caseStudyData,otherParameters,low_habitat_limit,up_habitat_limit, Fit_and_p, FVr_bestmemit, fitMaxVector, gen, evalCount, I_repeatEvals, FullStoreSol, FullStoreVal, FullStoreCount);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Solo por si el framework explota debido a que las corridas tengan 
% diferente nmero de iteraciones 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

[fmv_hs,fmv_ws] = size(fitMaxVectorAux);

fitMaxVector = nan(fmv_hs,I_maxFits);
for ii = 1:I_maxFits
    if ii<=fmv_ws
      fitMaxVector(1,ii) = fitMaxVectorAux(1,ii); 
    else
      fitMaxVector(1,ii) = fitMaxVectorAux(1,fmv_ws);   
    end
end

fitMaxVector(1,I_maxFits) = fitMaxVectorAux(1,fmv_ws);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

plotResults(fitMaxVector);
hold on

dlmwrite('rf.txt',fitMaxVector,'-append');
dlmwrite('rv.txt',FVr_bestmemit,'-append');

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
      
function codes = initializeCodes(FVr_minbound,FVr_maxbound,codeSize)
    ws = length(FVr_minbound);
    codes = nan(codeSize,ws);
    
    for jjj=1:ws
       range = FVr_maxbound(jjj) - FVr_minbound(jjj);
       for iii=1:codeSize 
          codes(iii,jjj) = FVr_minbound(jjj) + ( range / (codeSize - 1.0) * (iii-1.0));        
       end
       
       if (codes(iii,jjj)>FVr_maxbound(jjj))
           codes(codeSize,jjj) = FVr_maxbound(jjj);
       end
    end
end    

function enCodedPop = enCodePop(pop,codes,FVr_minbound,FVr_maxbound)
    [hs,ws] = size(pop);
    enCodedPop = nan(hs,ws);
    for iii=1:hs
       enCodedPop(iii,:) = enCode(pop(iii,:),codes,FVr_minbound,FVr_maxbound);
    end
end    
    
function enCoded = enCode(indiv,codes,FVr_minbound,FVr_maxbound)
    [hs,ws] = size(codes);
    enCoded = nan(1,ws);
    for jjj=1:ws
       if (FVr_minbound(jjj)==FVr_maxbound(jjj)) 
            enCoded(1,jjj)=1;
       else           
            range = (FVr_maxbound(jjj) - FVr_minbound(jjj))/hs;
            sum = FVr_minbound(jjj) + range;
            iii = 1;
            while (indiv(jjj) > sum) && (iii<hs) 
                  iii = iii + 1;
                  sum = sum + range;
            end
            enCoded(1,jjj)=iii;
       end
    end
end    
    
function deCodedPop = deCodePop(enCodedPop,codes)
    [hs,ws] = size(enCodedPop);
    deCodedPop = nan(hs,ws);
    for iii=1:hs
       deCodedPop(iii,:) = deCode(enCodedPop(iii,:),codes);
    end
end    
    
function indiv = deCode(enCoded,codes)
   [~,ws] = size(codes);
    indiv = nan(1,ws);   
    for jjj=1:ws
        pos = enCoded(1,jjj);
        v   = codes(pos,jjj);
        indiv(1,jjj)=v;
    end
end    

function NeighborsIndexes=getNeighborsIndexes(~,cellIndexW,~,I_StructureWidth,I_NeighborhoodType,I_NeighborhoodParam,I_indByCell)
    switch I_NeighborhoodType
        case 1 % left-right 2k-1
            NeighborsCellIndexes = getNeighborsCellIndexes_LeftRight(cellIndexW,I_StructureWidth,I_NeighborhoodParam);       
    end
    NeighborsIndexes = getCellIndividualLinearIndexes(NeighborsCellIndexes(1), I_indByCell);
    [~,ws] = size(NeighborsCellIndexes);
    for index = 2:ws
       M = getCellIndividualLinearIndexes(NeighborsCellIndexes(index), I_indByCell);
       NeighborsIndexes = [NeighborsIndexes M];
    end
end    

function SelectedIndividualIndexes = getSelectedIndividualIndexes(NeighborsIndexes,S_val,I_SelectionMethod,I_Selected)
    
    switch I_SelectionMethod
        case 1 % Trunc Selection
            SelectedIndividualIndexes = getTruncSelectedIndividualIndexes(NeighborsIndexes,S_val,I_Selected);   
        otherwise
            SelectedIndividualIndexes = [];
    end
end    
 
function [Px,Pxy] = getEstimadedParams(enCodedPop,SelectedIndividualIndexes,codes,I_EstimationParamsMethod,~)
    switch I_EstimationParamsMethod
        case 1    % 1 Bivariated
            %%%% Removed method. It is not relevant for RCED-UMDA
            % [Px,Pxy] = getEstimadedParamsBivariated(enCodedPop,SelectedIndividualIndexes,codes);
        otherwise % 0 Univariated
            [Px,Pxy] = getEstimadedParamsUnivariated(enCodedPop,SelectedIndividualIndexes,codes);
    end
end    
        
function [Px,Pxy] = getEstimadedParamsUnivariated(FM_pop,SelectedIndividualIndexes,codes)
    parents = FM_pop(SelectedIndividualIndexes, :);
    [chs,cws] = size(codes); 
    [phs,~] = size(parents);
    Pxy = nan;
    
    Px   = zeros(chs,cws);
    
    ivx = 1;
    while(ivx<=cws)
      
       id = 1;
       while (id <=phs)
          codeX = parents(id,ivx);
          Px(codeX,ivx) = Px(codeX,ivx)+1;              
          id = id + 1; 
       end
        
       ivx = ivx + 1; 
    end
end    
    
function [Dxgy,EdgesList,EdgesCount] = getEstimadedStructure(~,~,I_EstimationStructureMethod,~)
    switch I_EstimationStructureMethod
        case 1
            %%%% Removed method. It is not relevant for RCED-UMDA
            % [Dxgy,EdgesList,EdgesCount] = getEstimadedStructure_OrderPrecedenceMaxDependency(Px,Pxy,I_EstimationStructureParam); 
        otherwise
            Dxgy = zeros(1,1);
            EdgesList  = nan(1,1);
            EdgesCount = zeros(1,1);           
    end
end       
   
function [Px,Pxy] = getScaledParams(Px,Pxy,n,I_ScalerMethod,I_ScalerParam)
    switch I_ScalerMethod
        case 1 
           %%%% Removed method. It is not relevant for RCED-UMDA
           % [Px,Pxy] = getScaledParamsWithMinValue(Px,Pxy,n,I_ScalerParam);          
        case 2   
           [Px,Pxy] = getScaledParamsAddingValue(Px,Pxy,n,I_ScalerParam);
           
        otherwise    
                    
    end
end      
    
function [Px,Pxy] = getScaledParamsAddingValue(Px,Pxy,n,I_ScalerParam)
    [chs,~] = size(Px);  
    
    singleAdd = I_ScalerParam * n / chs;
    doubleAdd = I_ScalerParam * n / (chs*chs);
    
    Px  = singleAdd + Px;
    Pxy = doubleAdd + Pxy;
end    
    
function NewCellPop = sampleNewIndividualsFromModel(I_indByCell,Px,~,~,~,~,I_SamplingMethod,~)
   switch I_SamplingMethod
        case 1
           %%%% Removed method. It is not relevant for RCED-UMDA
           % NewCellPop = sampleNewIndividualsFrom_OrderPrecedenceMaxDependency(I_indByCell,Px,Pxy,Dxgy,EdgesList,EdgesCount);
        otherwise    
           NewCellPop = sampleNewIndividualsFromModelWithoutDependencies(I_indByCell,Px);         
   end
end      
    
function NewCellPop = sampleNewIndividualsFromModelWithoutDependencies(I_indByCell,Px)
    [~,vws] = size(Px);
     NewCellPop = rand(I_indByCell,vws);
    
     for ic = 1: I_indByCell            
        for iv = 1: vws
           NewCellPop(ic,iv) = rulete(NewCellPop(ic,iv), Px(:,iv));
        end
     end
end     
             
function value = rulete(rv, Counters) 
    len = length(Counters); 
    
    total = 0;
    for value=1:len
        total = total + Counters(value);
    end
    
    rv = rv * total;
    
    sum = 0;
    for value=1:len
        sum = sum + Counters(value);
        if (rv<=sum) 
            return;
        end
    end
    value = len;
end    
   
function NeighborsIndexes = getNeighborsCellIndexes_LeftRight(cellIndexW,I_StructureWidth,I_NeighborhoodParam)
    NeighborsIndexes = nan(1,2*I_NeighborhoodParam+1,'single');
    for index = 1:2*I_NeighborhoodParam+1
        NeighborsIndexes(index) = cellIndexW - I_NeighborhoodParam - 1 + index;
        if (NeighborsIndexes(index) < 1)
           NeighborsIndexes(index) = I_StructureWidth + NeighborsIndexes(index);
        end
        if (NeighborsIndexes(index) > I_StructureWidth)
           NeighborsIndexes(index) = NeighborsIndexes(index) - I_StructureWidth;
        end  
    end
end    

function CellIndividualLinearIndexes = getCellIndividualLinearIndexes(cellIndex,I_indByCell)
    CellIndividualLinearIndexes = nan(1,I_indByCell,'single');
    base = (cellIndex-1) * I_indByCell;
    for index = 1:I_indByCell
       CellIndividualLinearIndexes(index) = base + index;
    end
end    

function SelectedIndividualIndexes = getTruncSelectedIndividualIndexes(NeighborsIndexes,S_val,I_Selected)
    [~,niS] = size(NeighborsIndexes);
    NeighborsVal = nan(1,niS,'single');
    for index = 1:niS
        NeighborsVal(index)=S_val(NeighborsIndexes(index));
    end
    [~,I] = sort(NeighborsVal);

    SelectedIndividualIndexes = nan(1,I_Selected);
    for index = 1:I_Selected
        SelectedIndividualIndexes(index)=NeighborsIndexes(I(index));
    end
end    
 
function plotResults(Results)
    [~,ws] = size(Results);
    X=1:1:ws;
    Y=Results;
    plot(X,Y); 
    grid on
end    
    
% VECTORIZED THE CODE INSTEAD OF USING FOR
function pop=genpop(a,b,lowMatrix,upMatrix)
    pop=unifrnd(lowMatrix,upMatrix,a,b);
end

function [Val, evalCount] = my_feval(fnc,deCodedPop,caseStudyData, otherParameters, evalCount, I_repeatEvals)    
    switch I_repeatEvals
        case 0
           %%%% Removed method. It is not relevant for RCED-UMDA
           % [Val, evalCount] = my_feval_withStore(fnc,deCodedPop,caseStudyData, otherParameters, evalCount);
        case 2
           [Val, evalCount] = my_feval_withStoreForElitism(fnc,deCodedPop,caseStudyData, otherParameters, evalCount); 
        otherwise    
            [s,~] = size(deCodedPop);    
            evalCount = evalCount + s;
            [Val, ~] = feval(fnc,deCodedPop,caseStudyData, otherParameters);
            fprintf('evalCount: %d\n',evalCount)      
    end
end  

     
function index = searchSolution(sol, ws)
     index = 0; 
    
     for iii=1:FullStoreCount
                 
         flag = 1;
         for jjj=1:ws
            if FullStoreSol(iii,jjj)~=sol(jjj)
              flag = 0;
              break;
            end
         end
         
         if flag == 1
            index = iii;
            fprintf('*');
            break; 
         end         
     end
end             

end