clc;clear all;close all;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% IEEE CEC/GECCO 2023 competition: Competition on Evolutionary Computation in the Energy Domain: Operation and Planning Applications
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% ALGORITMH: CHAOTIC LEAVY HC2RCEDUMDA (CL_HC2RCEDUMDA)
% CHAOTIC LEAVY HC2RCEDUMDA
%This algorithm is the modified version of the HC2RCEDUMDA using CHAOTIC LEAVY distribution.
% The codes have been modified by the developers to propose CL_HC2RCEDUMDA
%% Developers: 
% Main author: Dharmesh A. Dabhi, POST DOCTORAL RESEARCHER, TECHNOLOGICAL UNIVERSITY DUBLIN,DUBLIN
% Co-authors: 
% Kartik S. Pandya,Independent Researcher
% Pratik Mochi, Assistant professor, M & V Patel Department of Electrical Engineering, CSPIT,CHARUSAT UNIVERSITY,CHANGA, Gujarat, INDIA
% ROHIT SALGOTRA, FACULTY OF PHYSICS & APPLIED COMPUTER SCIENCE, AGH UNIVERSITY OF SCIENCE & TECHNOLOGY, POLAND.
%% References:
% The codes of HC2RCEDUMDA available http://www.gecad.isep.ipp.pt/ERM-competitions/2021-2/
% The Codes of HC2RCEDUMDA were developed by the following team in IEEE PES GM 2021, IEEE %CEC 2021 & GECCO 2021 Joint competition 2021
% Authors: Ansel Y. Rodríguez González 1,2 (Team Leader) 
%          Samantha Barajas            1,3
%          Ramón Aranda                1,2
%          Yoan Martínez López         4
%          Julio Madera                4
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
addpath('functions') %Necessary functions to run the algorithms (encrypted)
tTotalTime=tic; % lets track total computational time
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% Select testbed
Select_testbed=2; 
%Testbed 1: Risk-based energy scheduling
%Testbed 2: Transmission expansion planning

DB=Select_testbed;
% 1: Case study testbed 1
% 2: Case study testbed 2

Select_algorithm=2;
%1: HyDE algorithm (test algorithm)
%2: Your algorithm

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Load MH parameters (e.g., get MH parameters from DEparameters.m file)
switch Select_algorithm
       case 1
          addpath('HyDE')
          algorithm='HyDE-Algorithm'; %'The participants should include their algorithm here'
          DEparameters %Function defined by the participant
          No_solutions=deParameters.I_NP; %Notice that some algorithms are limited to one individual
    case 2
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %Your algorithm can be inserted here
         %Your algorithm can be inserted here
        addpath('CL_HC2RCEDUMDA')
        algorithm='CL_HC2RCEDUMDA'; %'The participants should include their algorithm here'
        cl_HC2RCEDUMDAparameters %Function defined by the participant
        No_solutions=CL_HC2RCEDUMDAparameters.I_NP; %Notice that some algorithms are limited to one individual    
         
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
    otherwise
        fprintf(1,'No algorithm selected\n');
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Load Data base
%% Load Data base 
[caseStudyData, DB_name]=callDatabase(DB);
noRuns=20; %this can be changed but final results should be based on 20 trials

%% Label of the algorithm and the case study
Tag.algorithm=algorithm;
Tag.DB=DB_name;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Set other parameters
otherParameters =setOtherParameters(caseStudyData, No_solutions, Select_testbed);
otherParameters.one = setOtherParameters(caseStudyData,1, Select_testbed);
%% Set lower/upper bounds of variables 
[lowerB,upperB] = setVariablesBounds(caseStudyData,otherParameters, Select_testbed);
 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Call the MH for optimizationclear 
ResDB=struc([]);
for iRuns=1:noRuns %Number of trails
    tOpt=tic;
    rand('state',sum(noRuns*100*clock))% ensure stochastic indpt trials
    otherParameters.iRuns=iRuns;      
        
    switch Select_algorithm
        case 1
            [ResDB(iRuns).Fit_and_p, ...
            ResDB(iRuns).sol, ...
            ResDB(iRuns).fitVector]= ...                    
            HyDE(deParameters,caseStudyData,otherParameters,lowerB,upperB); 
        case 2
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            %Your algorithm can be inserted here
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
             [ResDB(iRuns).Fit_and_p, ...
                ResDB(iRuns).sol, ...
                ResDB(iRuns).fitVector]= ...
                CL_HC2RCEDUMDA(CL_HC2RCEDUMDAparameters,caseStudyData,otherParameters,lowerB,upperB);
    end
    
   ResDB(iRuns).tOpt=toc(tOpt); % time of each trial
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            %% Save the results and stats
        Save_results
        
end

%% End of MH Optimization
for j=1:noRuns
    lower_violations=ResDB(j).sol<lowerB;
    upper_violations=ResDB(j).sol>upperB;
    many_low(j)=sum(lower_violations);
    many_up(j)=sum(upper_violations);
    fprintf('lower_violations : %d .\n', many_low(j));
    fprintf('upper_violations : %d .\n', many_up(j));
end

tTotalTime=toc(tTotalTime); %Total time